from dockwrkr import (Command)

class Help(Command):

  def __init__(self):
    super(Help, self).__init__()
    self.autoInitCore = False
  
  def getUsage(self):
    return "dockwrkr help [command?]"

  def getHelpTitle(self):
    return "Print help for a specific command"

  def getShellOptions(self, optparser):
    return optparser

  def main(self):
    if len(self.args) > 0:
      parent = self.parent
      for commandName in self.args:
        command = parent.getCommand(commandName)
        command.initialize()
        parent = command
    else:
      command = self.parent
    return command.exitHelp()

  def getInputCommand(self):
    name = self.args[0]
    return name




import logging
from dockwrkr.monads import *
from dockwrkr import (Command)
import tabulate

logger = logging.getLogger(__name__)

class Status(Command):

  def getShellOptions(self, optparser):
    return optparser

  def getUsage(self):
    return "dockwrkr status [options] CONTAINER..."

  def getHelpTitle(self):
    return "Output the container status table"

  def main(self):
    containers = self.args
    return self.core.status(self.args) \
      .bind(self.tabulateStatus) \
      .catch(self.exitError) \
      .bind(logger.info)

  def tabulateStatus(self, containerStatuses):
    headers = ["NAME", "CONTAINER", "PID", "IP", "UPTIME", "EXIT"]
    tabulate.MIN_PADDING = 8
    table = tabulate.tabulate(containerStatuses, headers=headers, tablefmt="plain")
    return OK(table)



import unittest
import tempfile
import sys
import pprint
from dockwrkr.shell import Shell

class TestBase(unittest.TestCase):

  tempFiles = []
  tempDirs = []

  @classmethod
  def addTemporaryFile(cls):
    tempFile = tempfile.NamedTemporaryFile()
    cls.tempFiles.append(tempFile)
    return tempFile

  @classmethod
  def addTemporaryDir(cls):
    tempDir = tempfile.mkdtemp()
    cls.tempDirs.append(tempDir)
    return tempDir

  @classmethod
  def clearTemporaryDirs(cls):
    for x in cls.tempDirs:
      Shell.nukeDirectory(x)

  @classmethod
  def clearTemporaryFiles(cls):
    for x in cls.tempFiles:
      x.close()

  @staticmethod
  def raiser(e):
    raise e



from dockwrkr import (Command)

class Reset(Command):

  def __init__(self):
    super(Reset, self).__init__()
    self.autoInitCore = False

  def getShellOptions(self, optparser):
    optparser.add_option("-t","--time", dest="time", help="Seconds to wait before sending SIGKILL", default=10)
    return optparser

  def getUsage(self):
    return "dockwrkr reset [options]"

  def getHelpTitle(self):
    return "Reset container managed by dockwrkr (stop/remove)"

  def main(self):
    return self.core.reset(time=self.getOption('time')).catch(self.exitError)



from dockwrkr import (Command)
from dockwrkr.console import PassThroughParser

class Exec(Command):

  def getParserClass(self):
    return PassThroughParser

  def getShellOptions(self, optparser):
    optparser.add_option("-t","--tty", dest="tty", help="Allocate a pseudo-TTY", action="store_true")
    optparser.add_option("-i","--interactive", dest="interactive", help="Keep STDIN open even if not attached", action="store_true")
    optparser.add_option("--privileged", dest="privileged", help="Give extended privileges to the command", action="store_true")
    optparser.add_option("-d", "--detach", dest="detach", help="Run in the background", action="store_true")
    optparser.add_option("-u", "--user", dest="user", help="Username or UID")
    return optparser

  def getUsage(self):
    return "dockwrkr exec [options] CONTAINER"

  def getHelpTitle(self):
    return "Run a command in a running container"

  def main(self):
    containers = self.args
    if not len(self.args) > 0:
      return self.exitWithHelp("Please provide a container.")
    if not len(self.args) > 1:
      return self.exitWithHelp("Please provide a command.")
    container = self.args[0]
    command = self.args[1:]
    return self.core.excmd(
      container=container, 
      cmd=command,
      tty=self.getOption('tty'),
      user=self.getOption('user'), 
      detach=self.getOption('detach'),
      privileged=self.getOption('privileged'),
      interactive=self.getOption('interactive')
      ).catch(self.exitError)



import os
import sys
import logging
import re
from dockwrkr.monads import *
from dockwrkr.logs import *
from dockwrkr.exceptions import *
from dockwrkr.shell import Shell
from dockwrkr.utils import (readYAML, mergeDict, ensureList, dateToAgo, walkUpForFile, writeToFile, expandLocalPath)
import dockwrkr.docker as docker

logger = logging.getLogger(__name__)

class Core(object):

  def __init__(self):
    self.options = {}
    self.configFile = None
    self.initialized = False
    self.config = {}
    return

  def initialize(self):
    if self.initialized:
      return OK(None)
    return self.loadConfig().then(defer(self.setInitialized, b=True))

  def setInitialized(self, b):
    self.initialized = b

  def loadConfig(self):
    return self.readConfigFile() >> self.setConfig

  def findConfigFile(self):
    if self.configFile:
      return OK(self.configFile)

    configFile = walkUpForFile(os.getcwd(), "dockwrkr.yml")
    if not configFile:
      return Fail(ConfigFileNotFound("Could not locate config file: dockwrkr.yml"))
    self.configFile = configFile
    return OK(configFile)
 
  def readConfigFile(self):
    return self.findConfigFile().bind(lambda f: Try.attempt(readYAML, f))

  def setConfig(self, config):
    mergeDict(self.config, config)
    return OK(self)

  def getRegistries(self):
    regs = self.config.get('registries', {})
    return regs

  def getNetworks(self):
    networks = self.config.get('networks', {})
    return networks

  def getDefinedContainers(self):
    graph = []
    containers = self.config.get('containers')
    if not containers:
      containers = self.config
      self.legacyConfig = True

    for container in containers.keys():
      node = self.getContainerDependencies(container)
      graph.append( node )

    def resolveDependencies(node, resolved):
      for dep in node['deps']:
        depnode = self.getContainerDependencies(dep)
        if depnode['name'] not in resolved:
          resolveDependencies(depnode, resolved)
      if node['name'] not in resolved:
        resolved.append(node['name'])
    resolved = []
    for node in graph:
      resolveDependencies(node, resolved)

    return resolved

  def getContainerDependencies(self, container):
    node = {}
    node['name'] = container
    deps = []
    config = self.getContainerConfig(container)
    if "link" in config:
      for link in ensureList( config['link'] ):
        deps.append( link.partition(':')[0] )
    node['deps'] = deps
    return node

  def getBasePath(self):
    return os.path.dirname(self.configFile)

  def getContainerConfig(self, container):
    return self.config.get('containers', {}).get(container)

  def getContainerImage(self, container):
    conf = self.getContainerConfig(container)
    return conf.get('image', None)

  ### Commands ###

  def readOrderedContainers(self, containers=[]):
    defined = self.getDefinedContainers()
    missing = [x for x in containers if x not in defined]
    ordered = [x for x in defined if x in containers]
    if missing:
      return Fail(InvalidContainerError("Container '%s' not defined." % ' '.join(missing)))
    return OK(ordered)

  def create(self, containers=[], all=False):
    return self.__command(self.__create, containers=containers, all=all)

  def start(self, containers=[], all=False):
    return self.__command(self.__start, containers=containers, all=all)

  def stop(self, containers=[], all=False, time=docker.DOCKER_STOP_TIME):
    return self.__command(self.__stop, containers=containers, all=all, time=time)

  def remove(self, containers=[], all=False, time=docker.DOCKER_STOP_TIME, force=False):
    return self.__command(self.__remove, containers=containers, all=all, time=time, force=force)

  def restart(self, containers=[], all=False, time=docker.DOCKER_STOP_TIME):
    return self.__command(self.__restart, containers=containers, all=all, time=time)

  def stats(self, containers=[]):
    if not containers:
      containers = self.getDefinedContainers()
    return self.__command(self.__stats, containers=containers)

  def status(self, containers=[]):
    if not containers:
      containers = self.getDefinedContainers()
    return self.__readStates(containers) \
      .bind(self.__status, containers=containers)

  def reset(self, time=docker.DOCKER_STOP_TIME):
    managed = docker.readManagedContainers()
    if managed.isFail():
      return managed

    return managed \
      .bind(docker.filterExistingContainers) \
      .bind(docker.readContainersStatus) \
      .bind(self.__remove, containers=managed.getOK(), time=time, force=True)

  def pull(self, containers=[], all=False):
    if all:
      containers = self.getDefinedContainers()

    registries = self.getRegistries()

    def pullImage(container):
      image = self.getContainerImage(container)
      unpacked = docker.unpackImageString(image)
      registry = unpacked.get('registry')
      op = OK(None)
      if registry and registry in registries:
        op.then(defer(self.login, registry=registry))
      return op.then(defer(docker.pull, image=image)) \
        .then(dinfo("'%s' (%s) has been pulled." % (container, image)))

    return Try.sequence(map(pullImage, containers))

  def login(self, registry):
    registries = self.getRegistries()
    config = registries.get(registry)
    if not config:
      return Fail(InvalidRegistry("Invalid registry specified for login. It is not configured."))
    logger.info("Logging into registry: %s" % registry)
    return docker.login(registry, config.get('username'), config.get('password'), config.get('email'))
 
  def recreate(self, containers=[], all=False, time=docker.DOCKER_STOP_TIME):
    if all:
      containers = self.getDefinedContainers()
    return self.__readStates(containers) \
      .bind(self.__remove, containers=containers, force=True, time=time) \
      .then(defer(self.__readStates, containers=containers)) \
      .bind(self.__start, containers=containers)

  def excmd(self, container, cmd, tty=False, interactive=False, user=None, detach=None, privileged=None):
    return self.__readStates([container]) \
      .bind(self.__exec, container=container, cmd=cmd, tty=tty, interactive=interactive, user=user, detach=detach, privileged=privileged)

  def __stats(self, state, containers=[]):
    existing = [ x for x in containers if x in state ]
    return docker.stats(existing)
 
  def __exec(self, state, container, cmd, tty=False, interactive=False, user=None, detach=False, privileged=False):
    if container not in state:
      return Fail(InvalidContainerError("'%s' does not exist. Cannot execute command." % container))
    if not state[container].running:
      return Fail(InvalidContainerError("'%s' is not running. Cannot execute command." % container))

    return docker.execmd(container, cmd, tty, interactive, user, detach, privileged)
  
  def __command(self, func, containers=[], all=False, *args, **kwargs):
    if all:
      containers = self.getDefinedContainers()
    return self.__readStates(containers) \
      .bind(func, containers=containers, *args, **kwargs)

  def __readStates(self, containers):
    return self.readOrderedContainers(containers) \
      .bind(docker.filterExistingContainers) \
      .bind(docker.readContainersStatus) 

  def __status(self, state, containers=[]):
    table = []
    for container in containers:
      if container not in state:
        status = docker.ContainerStatus(container)
      else:
        status = state[container]

      row = [
        container, 
        status.getCol('cid'), 
        status.getCol('pid'), 
        status.getCol('ip'),
        dateToAgo(status.startedat) if status.startedat else "-",
        docker.getErrorLabel(status) if not status.running else "-"
      ]
      table.append(row)
    return OK(table) 

  def __create(self, state, containers=[]):
    ops = []
    for container in containers:
      if container not in state:
        op = docker.create(container, self.getContainerConfig(container), basePath=self.getBasePath(), networks=self.getNetworks()) \
          .then(dinfo("'%s' has been created." % container)) 
        ops.append(op)
      else:
        logger.warn("'%s' already exists." % container)
    return Try.sequence(ops)

  def __start(self, state, containers=[]):
    ops = []
    for container in containers:
      if container not in state:
        op = docker.create(container, self.getContainerConfig(container), basePath=self.getBasePath(),  networks=self.getNetworks()) \
          .then(defer(docker.start, container=container)) \
          .then(dinfo("'%s' has been created and started." % container))  \
          .then(defer(self.writePid, container=container))
        ops.append(op)
      else:
        if not state[container].running:
          op = docker.start(container) \
            .bind(dinfo("'%s' has been started." % container)) \
            .then(defer(self.writePid, container=container))
          ops.append(op)
        else:
          logger.warn("'%s' is already running." % container)
    return Try.sequence(ops)

  def __stop(self, state, containers=[], time=docker.DOCKER_STOP_TIME):
    ops = []
    for container in containers:
      if container not in state:
        logger.warn("Container '%s' does not exist." % container)
      else:
        if state[container].running:
          op = docker.stop(container, time) \
            .bind(dinfo("'%s' has been stopped." % container)) \
            .then(defer(self.clearPid, container=container))
        else:
          logger.warn("'%s' is not running." % container)
    return Try.sequence(ops)

  def __remove(self, state, containers=[], force=False, time=docker.DOCKER_STOP_TIME):
    logger.debug("REMOVE %s" % containers)
    ops = []
    for container in containers:
      if container in state:
        if state[container].running:
          if not force:
            logger.error("'%s' is running and 'force' was not specified." % container)
          else:
            op = docker.stop(container, time=time) \
              .then(defer(docker.remove, container=container)) \
              .bind(dinfo("'%s' has been stopped and removed." % container)) \
              .then(defer(self.clearPid, container=container))
            ops.append(op)
        else:    
          ops.append(docker.remove(container).bind(dinfo("'%s' has been removed." % container)))
    return Try.sequence(ops)

  def __restart(self, state, containers=[], time=docker.DOCKER_STOP_TIME):
    ops = []
    for container in containers:
      if container not in state:
        logger.error("'%s' does not exist." % container)
      else:
        if state[container].running:
          op = docker.stop(container, time=time) \
            .then(defer(docker.start, container=container)) \
            .bind(dinfo("'%s' has been restarted." % container)) \
            .then(defer(self.writePid, container=container))
          ops.append(op)
        else:
          ops.append(docker.start(container) \
            .bind(dinfo("'%s' has been started." % container))) \
            .then(defer(self.writePid, container=container))

    return Try.sequence(ops)

  def getPidsConf(self):
    return self.config.get('pids', {})
  
  def getPidsDir(self):
    dir = self.getPidsConf().get('dir', '/var/run/dockwrkr')
    return expandLocalPath(dir, self.getBasePath())

  def arePidsEnabled(self):
    return self.getPidsConf().get('enabled', False)

  def writePid(self, container):
    if not self.arePidsEnabled():
      return OK(None)

    dir = self.getPidsDir()
    pidfile = os.path.join(dir, "%s.pid" % (container))

    return Shell.makeDirectory(dir) \
      .then(defer(docker.readContainerPid, container=container)) \
      .bind(defer(Try.attempt, writeToFile, filename=pidfile))

  def clearPid(self, container):
    if not self.arePidsEnabled():
      return OK(None)

    dir = self.getPidsDir()
    pidfile = os.path.join(dir, "%s.pid" % (container))

    if os.path.isfile(pidfile):
      return Try.attempt(os.remove, pidfile)
    else:
      return OK(None)



import logging
from dockwrkr.monads import *
from dockwrkr.shell import Shell
from dockwrkr.utils import (ensureList, expandLocalPath, safeQuote)
from dockwrkr.exceptions import ShellCommandError, DockerError
import arrow

logger = logging.getLogger(__name__)

DOCKWRKR_LABEL_DOMAIN='ca.turbulent.dockwrkr'
DOCKWRKR_LABEL_DOMAIN_NETWORK='ca.turbulent.dockwrkr-network'

DOCKER_STOP_TIME = 10
DOCKER_CLIENT = "docker"

DOCKER_LIST_OPTIONS = [
  'add-host',
  'cap-add',
  'cap-drop',
  'device',
  'dns',
  'dns-opt',
  'dns-search',
  'env',
  'env-file',
  'expose',
  'group-add',
  'label',
  'label-file',
  'link',
  'log-opt',
  'lxc-conf',
  'publish',
  'security-opt',
  'ulimit',
  'volume',
  'volumes-from',
  'extra-flags',
]

DOCKER_MAPEQUAL_OPTIONS = [
  'env',
]

DOCKER_MAP_OPTIONS = [
  'add-host',
  'link',
  'volume',
  'publish',
]


DOCKER_SINGLE_OPTIONS = [
  'blkio-weight',
  'cpu-shares',
  'cpu-period',
  'cpu-quota',
  'cpuset-cpus',
  'cpuset-mems',
  'cgroup-parent',
  'cidfile',
  'entrypoint',
  'hostname',
  'ipc',
  'kernel-memory',
  'log-driver',
  'mac-address',
  'memory',
  'memory-reservation'
  'memory-swap',
  'memory-swapiness',
  'name',
  'net',
  'ip',
  'pid',
  'restart',
  'user',
  'uts',
  'stop-signal',
  'workdir',
]

DOCKER_BOOL_OPTIONS = [
  'disable-content-trust',
  'oom-kill-disable',
  'publish-all',
  'privileged',
  'read-only',
  'sig-proxy',
]

DOCKER_NETWORK_LIST_OPTIONS = [
  "gateway",
  "subnet",
  "ip-range",
  "label",
]

DOCKER_NETWORK_SINGLE_OPTIONS = [
  "driver",
  "ipam-driver",
]
DOCKER_NETWORK_MAP_OPTIONS = [
  ]

DOCKER_NETWORK_MAPEQUAL_OPTIONS =[
  "opt",
  "aux_address",
  "ipam-opt",
]
DOCKER_NETWORK_BOOL_OPTIONS = [
  "internal",
  "ipv6",
]

def dockerClient(cmd, params=""):
  return assertDockerVersion().then(defer(dockerCommand, cmd, params))

def dockerCommand(cmd, params="", shell=False, stream=False, cwd=None):
  return Shell.command("%s %s %s" % (DOCKER_CLIENT, cmd, params), shell=False) \
    .catch(onDockerError)

def onDockerError(err):
  return Fail(DockerError(
    message=err.stderr,
    code=err.code,
    stdout=err.stdout,
    stderr=err.stderr,
    cmd=err.cmd))

#-- API

def readManagedContainers():
  filters = "-q -a --filter 'label=%s.managed=1' --format '{{.Label \"%s.name\"}}'" % (DOCKWRKR_LABEL_DOMAIN, DOCKWRKR_LABEL_DOMAIN)
  return dockerCommand("ps", filters) \
    .bind(parseContainerList)

def filterExistingContainers(containers):
  return readManagedContainers() \
    .map(lambda l: filter(lambda x: x in containers, l))

def readContainerExists(container):
  filter = "-q -a --filter \"label=%s.name=%s\" --format '{{.Label \"%s.name\"}}'" % (DOCKWRKR_LABEL_DOMAIN, safeQuote(container), DOCKWRKR_LABEL_DOMAIN)
  return dockerCommand("ps", filter) \
    .bind(self.parseContainerList) \
    .map(self.__listToBool)

def readNetworkExists(network):
  net_filter = '-q --filter \"name=%s\"' % (network.keys()[0])
  if dockerCommand('network ls', net_filter) \
    .bind(parseContainerList) \
    .map(__listToBool).value:
    return Fail("Network Exists")
  else:
    return OK(network)

def readContainerRunning(container):
  inspect = "--format=\"{{.State.Running}}\" %s 2> /dev/null" % (safeQuote(container))
  return dockerCommand("inspect", inspect) \
    .bind(lambda r: OK(True) if r['stdout'] == "true" else OK(False))

def readContainerGhosted(container):
  inspect = "--format=\"{{ .State.Ghost }}\" %s" % (safeQuote(container))
  def ghosted(r):
    if r.out == "true":
      logger.warning("WARNING - '%s' is ghosted." % container)
      return OK(True)
    return OK(False)
  return dockerCommand("inspect", inspect) \
    .bind(ghosted)

def readContainersStatus(containers=[]):

  if not containers:
    return OK({})

  inspect = "-f '{{.Name}}|{{.Id}}|{{.Config.Image}}|{{.NetworkSettings.IPAddress}}|{{range $p, $conf := .NetworkSettings.Ports}}{{if $conf}}{{$p}}->{{(index $conf 0).HostPort}}{{end}} {{end}}|{{.State.Pid}}|{{.State.StartedAt}}|{{.State.Running}}|{{.State.ExitCode}}|{{.State.Error}}' %s" % (' '.join(containers))
  return dockerCommand("inspect", inspect) \
    .bind(parseContainerStatus)

def readContainerPid(container):
  inspect = "--format '{{.State.Pid}}' %s" % (safeQuote(container))
  return dockerCommand("inspect", inspect) \
    .bind(lambda r: OK(r['stdout'].strip()))

def create(container, config, basePath=None, networks=None):
  params = readCreateParameters(container, config, basePath=basePath, networks=networks)
  if params.isFail():
    return params
  return dockerCommand("create", params.getOK())

def createNetwork(network):
  params = readCreateNetworkParameters(network)
  if params.isFail():
    return params
  return dockerCommand("network create", params.getOK())

def start(container):
  return dockerCommand("start", container)

def stop(container, time=10):
  params = "-t %s %s" % (time, container)
  return dockerCommand("stop", params)
 
def remove(container, force=False):
  params = container
  if force:
    params = "-f %s" % container
  return dockerCommand("rm", params)

def pull(image):
  return Shell.call("%s %s %s" % (DOCKER_CLIENT, "pull", image)) \
    .catchError(ShellCommandError, defer(_pullLoginChain, image=image))

def _pullLoginChain(err, image):
  parts = unpackImageString(image)
  if parts.get('registry'):
    return login(parts.get('registry')).then(defer(pull, image=image))
  else:
    return Fail(err)
  
def unpackImageString(imageStr):
  parts = imageStr.split('/', 1)
  unpack = {}
  if len(parts) > 1:
    unpack['registry'] = parts[0]
    unpack['image'] = parts[1]
  else:
    unpack['registry'] = None
    unpack['image'] = imageStr
  return unpack

def execmd(container, cmd, tty=False, interactive=False, user=None, detach=None, privileged=None):
  opts = []
  if tty:
    opts.append('--tty')
  if interactive:
    opts.append('--interactive')
  if user:
    opts.append('--user %s' % user)
  if detach:
    opts.append('--detach')
  if privileged:
    opts.append('--privileged')

  parts = []
  parts.append(' '.join(opts))
  parts.append(container)
  parts.append(' '.join(cmd))

  return Shell.call("%s %s %s" % (DOCKER_CLIENT, "exec", ' '.join(parts))) 

def stats(containers=[]):
  opts = ['-a']
  
  parts = []
  parts.append(' '.join(opts))
  parts.append(' '.join(containers))

  return Shell.call("%s %s %s" % (DOCKER_CLIENT, "stats", ' '.join(parts))) 

def login(registry, username=None, password=None, email=None):
  opts = []
  if username:
    opts.append("-u %s" % username)
  if password:
    opts.append("-p %s" % password)
  if email:
    opts.append("-e %s" % email)

  return Shell.call("%s %s %s %s" % (DOCKER_CLIENT, "login", ' '.join(opts), registry))

def logout(registry):
  return dockerCommand("logout", "%s" % (registry))

def readCreateNetworkParameters(network):
  network_name = network.keys()[0]
  network_params = network[network_name]

  cmd = []
  cmd_opt = []
  cmd_list = []
  cmd_map = []

  for confkey, confval in network_params.iteritems():
    if confkey in DOCKER_NETWORK_MAP_OPTIONS or confkey in DOCKER_NETWORK_MAPEQUAL_OPTIONS:
      if isinstance(confval, dict):
        for ck, cv in confval.iteritems():
          if confkey in DOCKER_NETWORK_MAPEQUAL_OPTIONS:
            cmd_map.append("--%s=\"%s\"=\"%s\"" % (confkey, ck, safeQuote(cv)))
          else:
            cmd_map.append("--%s=%s:%s" % (confkey, ck, safeQuote(cv)))
        continue
    if confkey in DOCKER_NETWORK_SINGLE_OPTIONS:
      cmd_opt.append("--%s=%s" % (confkey, safeQuote(confval)))
    elif confkey in DOCKER_NETWORK_BOOL_OPTIONS:
      if confval is not None and (confval is True or confval.lower() in ["true", "yes"] or confval == 1):
        cmd_opt.append("--%s" % confkey)
    elif confkey in DOCKER_NETWORK_LIST_OPTIONS:
      if not isinstance(confval, (basestring, list, tuple)):
        return InvalidConfigError(
          "[%s] Malformed option '%s': Should be string, number or list." % (network_name, confkey))
      confval = ensureList(confval)
      for i, lv in enumerate(confval):
        cmd_list.append("--%s=%s" % (confkey, safeQuote(lv)))
    else:
      return InvalidConfigError("[%s] Unkown option '%s'." % (network_name, confkey))

  for part in cmd_opt:
    cmd.append(part)
  for part in cmd_list:
    cmd.append(part)
  for part in cmd_map:
    cmd.append(part)

  cmd.append("--label %s.name=\"%s\"" % (DOCKWRKR_LABEL_DOMAIN_NETWORK, network_name))
  cmd.append("--label %s.managed=1" % DOCKWRKR_LABEL_DOMAIN)
  cmd.append(network_name)

  cmdline = ""
  for i, part in enumerate(cmd):
    if i > 0:
      cmdline += " "
    cmdline += part

  return OK(cmdline)

def readCreateParameters(container, config, basePath=None, networks=None):

  cconf = config.copy()

  if 'net' in cconf:
    if networks and cconf['net'] in networks:
      readNetworkExists({ cconf['net'] : networks[cconf['net']] }) \
      .bind(createNetwork)

  cconf['name'] = container

  if 'image' not in cconf:
    return InvalidConfigError("[%s] Container has no 'image' defined." % (container))
  image = cconf['image']
  del cconf['image']

  command = None
  if 'command' in cconf:
    command = cconf['command']
    del cconf['command']

  extra_flags = []
  if 'extra-flags' in cconf:
    extra_flags = ensureList(cconf['extra-flags'])
    del cconf['extra-flags']

  cmd = []
  cmd_opt = []
  cmd_list = []
  cmd_map = []

  for confkey, confval in cconf.iteritems():
    if confkey in DOCKER_MAP_OPTIONS or confkey in DOCKER_MAPEQUAL_OPTIONS:
      if isinstance(confval, dict):
        for ck,cv in confval.iteritems():
          if confkey in DOCKER_MAPEQUAL_OPTIONS:
            cmd_map.append("--%s=%s=%s" % (confkey, ck, safeQuote(cv)))
          else:
            cmd_map.append("--%s=%s:%s" % (confkey, ck, safeQuote(cv)))
        continue
    if confkey == "volume":
      for ck in ensureList(confval):
        if ck.find(':'):
          (path,sep,path_map) = ck.partition(':')
          path = expandLocalPath(path, basePath=basePath)
          cmd_map.append("--%s=%s:%s" % (confkey, safeQuote(path), safeQuote(path_map)))
      continue
    if confkey in DOCKER_SINGLE_OPTIONS:
      cmd_opt.append("--%s=%s" % (confkey, safeQuote(confval)))
    elif confkey in DOCKER_BOOL_OPTIONS:
      if confval is not None and (confval is True or confval.lower() in ["true","yes"] or confval == 1):
        cmd_opt.append("--%s=%s" % (confkey, "true"))
      elif confval is not None and (confval is False or confval.lower() in ["false","no"] or confval == 0):
        cmd_opt.append("--%s=%s" % (confkey, "false"))
      else:
        return InvalidConfigError("[%s] Invalid value for option '%s' : Must be true or false." % (container, confkey))
    elif confkey in DOCKER_LIST_OPTIONS:
      if not isinstance(confval, (basestring, list, tuple)):
        return InvalidConfigError("[%s] Malformed option '%s': Should be string, number or list." % (container, confkey))
      confval = ensureList(confval)
      for i, lv in enumerate(confval):
        cmd_list.append("--%s=%s" % (confkey, safeQuote(lv)))
    else:
        return InvalidConfigError("[%s] Unkown option '%s'." % (container, confkey))

  for part in cmd_opt:
    cmd.append(part)
  for part in cmd_list:
    cmd.append(part)
  for part in cmd_map:
    cmd.append(part)
  for extra in extra_flags:
    cmd.append(extra)

  cmd.append("--label %s.name=\"%s\"" % (DOCKWRKR_LABEL_DOMAIN, container))
  cmd.append("--label %s.managed=1" % DOCKWRKR_LABEL_DOMAIN)
  cmd.append(image)

  if command:
    cmd.append("%s" % command)

  cmdline = ""
  for i, part in enumerate(cmd):
    if i > 0:
     cmdline += " "
    cmdline += part
    #if i != len(cmd)-1:
    #  cmdline += "  \\"
    #cmdline += "\n"

  return OK(cmdline)

#---- Parse command output

def parseContainerList(containers):
  return OK(containers['stdout'].strip().splitlines())

class ContainerStatus(object):
  def __init__(self, name):
    self.name = name
    self.cid = None
    self.image = None
    self.pid = None
    self.ip = None
    self.ports = None
    self.startedat = None
    self.running = None
    self.exitcode = None
    self.exiterr = None

  @staticmethod
  def fromStatusLine(line):
    parts = line.split('|')
    parts += [None] * (10 - len(parts))

    name = parts[0]
    cid = parts[1]
    image = parts[2]
    ip = parts[3]
    ports = parts[4]
    pid = parts[5]
    startedat = parts[6]
    running = parts[7]
    exitcode = parts[8]
    exiterr = parts[9]

    name = name[1:]
    status = ContainerStatus(name)
    status.cid = cid[0:12] if cid else None
    status.image = image
    status.ip = ip if ip else None
    status.pid = int(pid) if pid and pid != "0" else None
    status.running = True if running == 'true' else False
    status.startedat = arrow.get(startedat).timestamp if startedat and status.running else None
    status.exitcode = int(float(exitcode)) if exitcode else None
    status.exiterr = exiterr
    return status

  def getCol(self, field):
    if hasattr(self, field):
      v = getattr(self, field)
      return v if v is not None else "-"
    else:
      return "-"
     
def parseContainerStatus(inspect):
  statuses = {}
  for line in inspect['stdout'].strip().splitlines():
    status = ContainerStatus.fromStatusLine(line)
    statuses[status.name] = status
  return OK(statuses)

#-- Transformations

def getErrorLabel(status):
  errcode = status.exitcode
  err = status.exiterr
  if errcode == -1:
    return "%s" % err
  elif errcode == 1:
    return "General error"
  elif errcode == 126:
    return "Command invoked cannot execute"
  elif errcode == 127:
    return "Command not found"
  elif errcode == 137:
    return "SIGKILL received"
  elif errcode == 143:
    return "SIGTERM received"
  elif errcode == 0:
    return "-"
  elif errcode is not None or err is not None:
    return "Exit code %s: %s" % (errcode, err)
  else:
    return "-"

def __listToBool(l):
  return len(l) > 0



from dockwrkr import (Command)

class Remove(Command):

  def getShellOptions(self, optparser):
    optparser.add_option("-a","--all", dest="allc", help="Apply to defined containers", default=False, action="store_true")
    optparser.add_option("-f","--force", dest="force", help="Stop running containers", default=False, action="store_true")
    optparser.add_option("-t","--time", dest="time", help="Seconds to wait before sending SIGKILL", default=10)
    return optparser

  def getUsage(self):
    return "dockwrkr remove [options] CONTAINER..."

  def getHelpTitle(self):
    return "Remove the specified container(s)"

  def main(self):
    containers = self.args
    if not len(self.args) > 0 and not self.options.allc:
      return self.exitWithHelp("Please provide a container or use -a for all containers.")
    return self.core.remove(self.args, all=self.getOption('allc'), force=self.getOption('force'), time=self.getOption('time')) \
      .catch(self.exitError)



import logging
import functools
from functools import (wraps)
from functools import partial

class Monad(object):

  def bind(self, mf):
    raise NotImplementedError

  def map(self, mf):
    raise NotImplementedError

  def __rshift__(self, f):
    return self.bind(f)

  def __ilshift__(self, f):
    return self.bind(f)

#----------- EITHER

class Maybe(Monad):

  @staticmethod
  def of(value=None):
    if(value is None):
      return Nothing()
    else:
      return Just(value)

  def get(self):
    raise NotImplementedError

  def isNothing(self):
    if isinstance(self, Nothing):
      return True 

  def bind(self, mf):
    if self.isNothing():
      return self
    return mf(self.get())

  def map(self, f):
    if self.isNothing(): 
      return self
    return Just(f(self.get()))

  def getOrElse(self, default):
    return default if self.isNothing() else self.get()
    
class Just(Maybe):
  value = None

  def __init__(self, value):
    self.value = value

  def get(self):
    return self.value

  def __repr__(self):
        return "Just(%r)" % self.value

  def __eq__(self, other):
    if isinstance(other, Just):
      return self.value == other.value
    return False

class Nothing(Maybe):
  def get(self):
    raise NoValue

  def __repr__(self):
        return "Nothing"

  def __eq__(self, other):
    return isinstance(other, Nothing)

#----------- EITHER

class Either(Monad):

  def isLeft(self):
    return isinstance(self, Left)
       
  def isRight(self):
    return isinstance(self, Right)

  def getLeft(self):
    raise NotImplementedError

  def getRight(self):
    raise NotImplementedError

  def bind(self, mf):
    if self.isLeft():
      return self
    return mf(self.getRight())
  
  def map(self, mf):
    if self.isLeft():
      return self
    return Right(mf(self.getRight())) 

class Left(Either):
  value = None
  def __init__(self, value):
    self.value = value

  def getLeft(self):
    return self.value

  def __repr__(self):
        return "Left(%r)" % self.value

  def __eq__(self, other):
    if isinstance(other, self.__class__):
      return self.value == other.value
    return False
 
  
class Right(Either):
  value = None
  def __init__(self, value):
    self.value = value

  def getRight(self):
    return self.value

  def __repr__(self):
        return "Right(%r)" % self.value

  def __eq__(self, other):
    if isinstance(other, self.__class__):
      return self.value == other.value
    return False

#----------- TRY

class Try(Monad):

  @staticmethod
  def of(value):
    return OK(value)

  def isFail(self):
    return isinstance(self, Fail)
    
  def isOK(self):
    return isinstance(self, OK)

  def getError(self):
    raise NotImplementedError

  def getOK(self):
    raise NotImplementedError

  def getOrElse(self, default):
    return default if self.isFail() else self.getOK()

  def bind(self, mf, *args, **kwargs):
    if self.isFail():
      return self
    return mf(self.getOK(), *args, **kwargs)

  def map(self, f):
    if self.isFail():
      return self
    return OK(f(self.getOK()))

  def thenIfTrue(self, f):
    return self.thenIfBool(f, lambda: OK(False))

  def thenIfFalse(self, f):
    return self.thenIfBool(lambda: OK(True), f)

  def thenIfNone(self, f):
    if self.isFail():
      return self
    if self.getOK() is None:
      return self.then(f)
    else:
      return self
        
  def thenIfBool(self, fTrue, fFalse):
    if self.isFail():
      return self
    if self.getOK() is True:
      return self.then(fTrue)
    else:
      return self.then(fFalse)

  def then(self, okF=None, failF=None):
    if self.isFail(): 
      if failF:
        r = failF(self.getError())
        if isinstance(r, Try):
          return r
        return self
      else:
        return self
    else:
      if okF:
        r = okF()
        if isinstance(r, Try):
          return r
        return OK(r)
      else:
        return self

  def catch(self, f):
    return self.then(None, f)  

  def catchError(self, err, f):
    def catcher(x):
      if isinstance(x, err):
        return f(self.getError())
    return self.then(None, catcher)
      
  def bindIfTrue(self, f):
    return self.bindIf(f, lambda x: OK(False))

  def bindIfFalse(self, f):
    return self.bindIf(lambda x: OK(True), f)

  def bindIf(self, fTrue, fFalse):
    if self.isFail():
      return self
    if self.getOK() is True:
      return self.bind(fTrue)
    else:
      return self.bind(fFalse)

  def mapM(self, mf):
    def mapper(xs):
      return mapM(self, mf, xs)
    return self.bind(mapper)

  def mapM_(self, mf):
    def mapper(xs):
      return mapM_(self, mf, xs)
    return self.bind(mapper)

  @staticmethod
  def compose(*funcs):
    def func(x):
      return fold(lambda acc, f: acc.bind(f), funcs, Try.of(x))
    return func

  @staticmethod
  def sequence(monads):
    ''' Fold a list of monads into a monad containing the list of values '''
    return reduce(lambda acc, mv: unshiftM(Try, acc, mv), reversed(monads), Try.of([]))

  @staticmethod
  def attemptDeferred(f, expect=Exception):
    def tryer(*args, **kwargs):
      try:
        return OK(f(*args, **kwargs))
      except expect as e:
        return Fail(e)
    return tryer

  @staticmethod 
  def attempt(f, *args, **kwargs):
    return Try.attemptDeferred(f)(*args, **kwargs)

  @staticmethod 
  def raiseError(self, err=Exception):
    raise(err)

  def join(self):
    return self if self.isFail() else self.getOK()

class OK(Try):
  value = None
  def __init__(self, value):
    self.value = value

  def getOK(self):
    return self.value

  def __repr__(self):
        return "OK(%r)" % self.value

  def __bool__(self):
    return True
 
  def __eq__(self, other):
    if isinstance(other, self.__class__):
      return self.value == other.value
    return False

class Fail(Try):
  value = None
  def __init__(self, value):
    self.value = value

  def getError(self):
    return self.value

  def __bool__(self):
    return False

  def __repr__(self):
        return "Fail(%r)" % self.value

  def __eq__(self, other):
    if isinstance(other, self.__class__):
      return self.value == other.value
    return False

def fold(f, xs, init=None):
  return functools.reduce(f, xs, init)

def compose(*funcs): 
  def func(x):
    return fold(lambda acc, f: f(acc), reversed(funcs), x)
  return func

def unshiftM(monad, monads, mv): 
  ''' Prepend a monadic value to the list value of a monad '''
  return monads.bind(lambda xs: mv.bind(lambda x: monad.of( [x] + xs)))

def sequence(monad, monads):
  ''' Fold a list of monads into a monad containing the list of values '''
  return reduce(lambda acc, mv: unshiftM(monad, acc, mv), reversed(monads), monad.of([]))

def mapM(monad, mf, xs):
  ''' Map a monadic function over a list of values, lifting them into monadic context and convert the results to a monad containing a list of the values. '''
  mapped = map(mf, xs)
  sequenced = sequence(monad, mapped)
  return sequenced

def mapM_(monad, mf, xs):
  ''' Map a monadic  action to a structure, evaluate from left to right and ignore results. '''
  map(mf, xs)
  return monad.of(xs)

def defer(f, *args, **kwargs):
  fargs = args
  fkwargs = kwargs
  return partial(f, *fargs, **fkwargs)

def failWith(exception):
  return (lambda *x: Fail(exception))

def chainSelf(self, *_, **__):
  return self

def flatten(xs):
  return [ x for sl in xs for x in sl]




class DockwrkrError(Exception):
  '''
  Base class to all dockwrkr exceptions
  '''
  errorLabel = ""
  def __init__(self, message=''):
    super(DockwrkrError, self).__init__(message)
    self.errorLabel = message

class ConfigFileNotFound(DockwrkrError):
  ''' Config file not found '''

class ConfigSyntaxError(DockwrkrError):
  ''' Config syntax error '''

class FileSystemError(DockwrkrError):
  ''' File system exception '''

class FileDoesNotExist(DockwrkrError):
  ''' File error exception '''

class ShellCommandError(DockwrkrError):
  def __init__(self, code=None, message='', stdout='', stderr='', cmd=''):
    super(ShellCommandError, self).__init__(message)
    self.code = code
    self.stdout = stdout
    self.stderr = stderr
    self.cmd = cmd

class UserInterruptError(ShellCommandError):
  ''' Raised when the user interrupts a process '''

class InvalidCommandError(DockwrkrError):
  '''
  Raised when an invalid command is passed to CLI
  '''
class InvalidOptionError(DockwrkrError):
  '''
  Raised when an invalid option is passed to CLI
  '''

class InvalidConfigError(DockwrkrError):
  '''
  Raised on an invalid configuration is found
  '''

class InvalidContainerError(DockwrkrError):
  ''' Invalid container was specified '''

class DockerError(ShellCommandError):
  '''
  Raised when a docker error is encountered
  '''



import os, errno
import sys
import logging
import collections
import shutil
import hashlib
from time import time
from datetime import datetime
from pkg_resources import Requirement, resource_filename, require
import yaml

try:  # py3
    from shlex import quote
except ImportError:  # py2
    from pipes import quote

from dockwrkr.exceptions import (
  ConfigSyntaxError,
  FileSystemError,
  FileDoesNotExist
)

logger = logging.getLogger(__name__)

_yaml_mapping_tag = yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG

def _dict_representer(dumper, data):
    return dumper.represent_dict(data.iteritems())

def _dict_constructor(loader, node):
    return collections.OrderedDict(loader.construct_pairs(node))

yaml.add_representer(collections.OrderedDict, _dict_representer)
yaml.add_constructor(_yaml_mapping_tag, _dict_constructor)

def getPackageVersion():
  return require('dockwrkr')[0].version

def ensureList(v):
  if v:
    return [ x for x in (v if isinstance(v, (list, tuple)) else [v]) ]
  else:
    return []


def expandLocalPath(path, basePath):
  if not os.path.isabs(path):
    path = os.path.normpath(path)
    path = os.path.join(basePath, path)
  return os.path.abspath(os.path.expanduser(os.path.realpath(path)))

def walkUpForFile(root, findfile):
  lastRoot = root
  while root:
    for base, dirs, files in os.walk(root):
      del dirs[0:len(dirs)]
      if findfile in files:
        return os.path.join(base, findfile)
    lastRoot = root
    root = os.path.dirname(lastRoot)
    if root == lastRoot:
      break
 
def dateToAgo(time=False):
  """
  Get a datetime object or a int() Epoch timestamp and return a
  pretty string like 'an hour ago', 'Yesterday', '3 months ago',
  'just now', etc
  """
  now = datetime.now()
  if type(time) is int:
    diff = now - datetime.fromtimestamp(time)
  elif isinstance(time,datetime):
    diff = now - time
  elif not time:
    diff = now - now
  second_diff = diff.seconds
  day_diff = diff.days

  if day_diff < 0:
    return ''

  if day_diff == 0:
    if second_diff < 10:
      return "just now"
    if second_diff < 60:
      return str(second_diff) + " seconds ago"
    if second_diff < 120:
      return "a minute ago"
    if second_diff < 3600:
      return str(second_diff / 60) + " minutes ago"
    if second_diff < 7200:
      return "an hour ago"
    if second_diff < 86400:
      return str(second_diff / 3600) + " hours ago"
  if day_diff == 1:
    return "Yesterday"
  if day_diff < 7:
    return str(day_diff) + " days ago"
  if day_diff < 31:
    return str(day_diff / 7) + " weeks ago"
  if day_diff < 365:
    return str(day_diff / 30) + " months ago"
  return str(day_diff / 365) + " years ago"


def writeToFile(data, filename):
  try:
    with open(filename, "wb") as fh:
      fh.write(data)
  except Exception as err:
    raise FileSystemError("Failed to write to %s : %s" % (filename, err))
 
def writeYAML(filename, data):
  try:
    with open(filename, "w") as fileh:
      fileh.write(yaml.dump(data, default_flow_style=False))
  except Exception as err:
    raise FileSystemError("Failed to write to %s : %s" % (filename, err))

def readYAML(filename):
  try:
    stream = open(filename, "r")
    contents = yaml.load(stream)
    return contents
  except yaml.YAMLError, exc:
    msg = "Syntax error in file %s"
    if hasattr(exc, 'problem_mark'):
      mark = exc.problem_mark
      msg += " Error position: (%s:%s)" % (mark.line+1, mark.column+1)
    raise ConfigSyntaxError(msg)
  except Exception as err:
    raise FileSystemError("Failed to read configuration file %s : %s" % (filename, err))

def humanReadableBytes(num):
  for x in ['bytes','KB','MB','GB']:
    if num < 1024.0:
      return "%3.1f%s" % (num, x)
    num /= 1024.0
    return "%3.1f%s" % (num, 'TB')

def sha1sum(filename):
  sha = hashlib.sha1()
  bufsize = 65536
  with open(filename, 'rb') as fd:
    while True:
      data = fd.read(bufsize)
      if not data:
        break
      sha.update(data)
  return sha.hexdigest()

def pathComponents(path):
  folders = []
  path = os.path.normpath(path)
  while 1:
    path, folder = os.path.split(path)
  
    if folder != "":
      folders.append(folder)
    else:
      if path != "":
        folders.append(path)
      break

  folders.reverse()  
  return folders

def mergeDict(a, b, path=None):
  if path is None: path = []
  for key in b:
    if key in a:
      if isinstance(a[key], dict) and isinstance(b[key], dict):
        mergeDict(a[key], b[key], path + [str(key)])
      elif a[key] == b[key]:
        pass # same leaf value
      else:
        a[key] = b[key]
    else:
      a[key] = b[key]
  return a

def safeQuote(string):
  return quote(str(string))
 



import unittest
import os
import tempfile
import tests
from contextlib import contextmanager

from dockwrkr.core import Core
from dockwrkr.shell import Shell
from dockwrkr.monads import *

class TestCore(tests.TestBase):

  def setUp(self):
    self.basePath = self.addTemporaryDir()

  def tearDown(self):
    ''' tearDown the test '''
    if self.basePath:
      Shell.nukeDirectory(self.basePath).catch(TestCore.raiser)

  def setUpDefaultCore(self):
    self.core = Core()
    return self.core

  def testInitialize(self):
    Shell.call("cp %s %s" % ("tests/dockwrkr-1.yml", os.path.join(self.basePath, "dockwrkr.yml")))
    with self.pushd(self.basePath):
      core = Core()
      self.assertIsInstance(core.initialize(), OK)
  
      self.assertIsInstance(core.reset(time=0), OK)
      self.assertIsInstance(core.start(containers=['hello1']), OK)
      self.assertIsInstance(core.start(containers=['hello2']), OK)
      self.assertIsInstance(core.start(containers=['hello3']), OK)
      self.assertIsInstance(core.stop(all=True, time=0), OK)
      self.assertIsInstance(core.start(all=True), OK)
      self.assertIsInstance(core.pull(all=True), OK)
      self.assertIsInstance(core.excmd(container='hello1', cmd=['ls','-al']), OK)
      self.assertIsInstance(core.reset(time=0), OK)

  def testPids(self):
    Shell.call("cp %s %s" % ("tests/dockwrkr-2.yml", os.path.join(self.basePath, "dockwrkr.yml")))
    with self.pushd(self.basePath):
      core = Core()
      self.assertIsInstance(core.initialize(), OK)
 
      pdir = os.path.join(self.basePath, "pids")

      self.assertIsInstance(core.reset(time=0), OK)
      self.assertIsInstance(core.start(containers=['hello1']), OK)
      self.assertTrue(os.path.isfile(os.path.join(pdir, "hello1.pid")))
      self.assertIsInstance(core.start(containers=['hello2']), OK)
      self.assertTrue(os.path.isfile(os.path.join(pdir, "hello2.pid")))
      self.assertIsInstance(core.start(containers=['hello3']), OK)
      self.assertTrue(os.path.isfile(os.path.join(pdir, "hello3.pid")))
      self.assertIsInstance(core.stop(all=True, time=0), OK)
      self.assertFalse(os.path.isfile(os.path.join(pdir, "hello1.pid")))
      self.assertFalse(os.path.isfile(os.path.join(pdir, "hello2.pid")))
      self.assertFalse(os.path.isfile(os.path.join(pdir, "hello3.pid")))

  def testRes(self):
    Shell.call("cp %s %s" % ("tests/dockwrkr-2.yml", os.path.join(self.basePath, "dockwrkr.yml")))
    with self.pushd(self.basePath):
      core = Core()
      self.assertIsInstance(core.initialize(), OK)
 
      self.assertIsInstance(core.reset(time=0), OK)
      self.assertIsInstance(core.start(containers=['hello1']), OK)
      self.assertIsInstance(core.restart(containers=['hello1'], time=0), OK)
      self.assertIsInstance(core.recreate(containers=['hello1'], time=0), OK)
      self.assertIsInstance(core.reset(time=0), OK)

  @contextmanager
  def pushd(self, newDir):
    previousDir = os.getcwd()
    os.chdir(newDir)
    yield
    os.chdir(previousDir)



from __future__ import absolute_import

import sys
import os
import re
import logging
import pkgutil
import importlib
import traceback
from collections import OrderedDict
from importlib import import_module
import optparse
from getpass import getpass
from dockwrkr.exceptions import InvalidCommandError

logger = logging.getLogger(__name__)

class Parser(optparse.OptionParser):
  pass
  def format_epilog(self, formatter):
    return "\n"+self.epilog if self.epilog else ""

class PassThroughParser(Parser):
  def _process_long_opt(self, rargs, values):
    try:
      optparse.OptionParser._process_long_opt(self, rargs, values)
    except optparse.BadOptionError, err:
      self.largs.append(err.opt_str)

  def _process_short_opts(self, rargs, values):
    try:
      optparse.OptionParser._process_short_opts(self, rargs, values)
    except optparse.BadOptionError, err:
      self.largs.append(err.opt_str)

class CLI(object):
  def __init__(self, assumeYes=False):
    self.assumeYes = assumeYes
    self.args = None
    self.options = None
    self.input = None
    self.parser = None
    self.name = None
    self.parent = None
    self.autoInitCore = True
  
  def getHelp(self):
    """Returns the help description for this particular command"""
    return self.getParser().format_help()

  def getHelpTitle(self):
    return ""

  def getHelpDetails(self):
    return ""

  def getUsage(self):
    return "command: USAGE [options]"

  def getParserClass(self):
    return Parser

  def getParser(self, interspersed=True):
    usage = self.getUsage()
    parserClass = self.getParserClass()
    parser = parserClass(usage=usage, conflict_handler="resolve", add_help_option=False, description=self.getHelpTitle(), epilog=self.getHelpDetails())
    self.getShellOptions(parser)
    if interspersed:
      parser.enable_interspersed_args()
    else:
      parser.disable_interspersed_args()
    return parser

  def parseShellInput(self, interspersed=True):
    """Return the options and arguments for this command as a tuple"""
    usage = self.getUsage()
    parser = self.getParser(interspersed)
    (opts, args) = parser.parse_args(self.input)
    return (opts, args)

  def getShellOptions(self, optparser):
    return optparser

  def execute(self, args=None):
    self.input = args
    (self.options, self.args) = self.parseShellInput()
    return self.main()

  def initialize(self):
    pass

  def main(self):
    pass

  def getOptions(self):
    return self.options

  def hasOption(self, key):
    if self.options and hasattr(self.options, key):
      return True

  def getOption(self, key):
    if self.hasOption(key):
      return getattr(self.options, key, None)
    return None

  def getArgs(self):
    return self.args

  def hasArg(self, argno):
    if len(self.args) < argno + 1:
      return False
    return True

  def getArg(self, argno):
    if self.hasArg(argno):
      return self.args[argno]
    return None

  def read(self, msg=""):
    return raw_input(msg)

  def readln(self, msg=""):
    return raw_input(msg + "\n")

  def readpwd(self, msg=""):
    return getpass(msg)

  def readpwdln(self, msg=""):
    return getpass(msg + "\n")

  def ask(self, msg):
    if self.assumeYes:
      return True
    print(msg)
    res = raw_input('Proceed? [N/y] ')
    if not res.lower().startswith('y'):
      return False
    print('... proceeding')
    return True

  def exitOK(self, msg=None):
    if msg:
      logger.info(msg)
    sys.exit(0)

  def exitHelp(self, msg=None, code=1):
    logger.info(self.getHelp())
    logger.info("")
    if msg:
      logger.info(msg)
    sys.exit(1)

  def exitError(self, msg=None, code=1):
    if msg and isinstance(msg, Exception):
      logger.error("%s (%s)" % (msg.message, type(msg).__name__))
    elif msg:
      logger.error(msg)
    sys.exit(1)


class Program(CLI):
  def __init__(self):
    super(Program, self).__init__()
    self.cmdString = None
    self.commandInput = None
    self.command = None
    self.core = None
    self.commands = OrderedDict()

  def setupCommands(self):
    pass
    
  def getCommands(self):
    k = self.commands.keys()
    v = map(lambda x: self.getCommand(x), k)
    return OrderedDict(zip(k, v))

  def getCommand(self, commandName):
    if commandName not in self.commands:
      raise InvalidCommandError("Invalid command '%s' specified" % commandName)
 
    commandModule = self.commands[commandName]
    module_ = importlib.import_module(commandModule)
    class_ = getattr(module_, commandName.title())
    command = class_()
    command.parent = self
    command.name = commandName
    return command

  def addCommand(self, commandName, commandModule):
    if commandName in self.commands:
      raise ValueError("Command '%s' is already registered." % commandName)
    self.commands[commandName] = commandModule
 
  def removeCommand(self, commandName):
    if commandName not in self.commands:
      raise ValueError("Command '%s' is not registered." % commandName)
    self.commands.pop(commandName, None)

  def getHelpDetails(self):
    helpUsage = "Commands:\n\n"
    for name, command in self.getCommands().iteritems():
      helpUsage += "  %-20s%s\n" % (name, command.getHelpTitle())
    return helpUsage

  def setupEnv(self):
    """Setup the Environment """

  def setupLogging(self):
    if self.getOption('debug'):
      logging.getLogger('dockwrkr').handlers[0].setLevel(logging.DEBUG)

  def execute(self, args=None):
    self.input = args
    (self.options, self.args) = self.parseShellInput(False)
    self.main()

  def initialize(self):  
    self.setupLogging()
    self.setupEnv()
    self.setupCommands()

  def main(self):
    try:
      self.initialize()

      if not len(self.args) > 0:
        self.exitError("Please provide a command.\n\nUse 'help' for available commands.")
  
      self.cmdString = self.args[0]
      self.commandInput = self.args[1:]

      if self.cmdString not in self.commands:
        return self.exitError("Invalid command '%s' specified for '%s'.\n\nUse 'help %s' for available commands." % (self.cmdString, self.name, self.name))

      self.runCommand(self.cmdString)
    except Exception as err:
      logger.error(traceback.format_exc())
      self.exitError("Error running command %s: %s" % (self.cmdString, err), code=2)

  def runCommand(self, commandName):
    command = self.getCommand(commandName)
    self.command = self.initCommand(command)
    self.command.execute(self.commandInput)

class Command(CLI):
  def __init__(self):
    super(Command, self).__init__()

class SubProgram(Program):
  def initialize(self):
    self.setupCommands()



from __future__ import absolute_import
import os
import sys
import logging
from dockwrkr import (Program, Core)
from dockwrkr.utils import getPackageVersion
import dockwrkr.logs

class DockwrkrCLI(Program):
  """Dockwrkr CLI command"""

  def __init__(self):
    super(DockwrkrCLI, self).__init__()
    self.name = 'dockwrkr'

  def setupCommands(self):
    self.addCommand('help', 'dockwrkr.command.help')
    self.addCommand('create', 'dockwrkr.command.create')
    self.addCommand('start', 'dockwrkr.command.start')
    self.addCommand('stop', 'dockwrkr.command.stop')
    self.addCommand('remove', 'dockwrkr.command.remove')
    self.addCommand('restart', 'dockwrkr.command.restart')
    self.addCommand('recreate', 'dockwrkr.command.recreate')
    self.addCommand('reset', 'dockwrkr.command.reset')
    self.addCommand('pull', 'dockwrkr.command.pull')
    self.addCommand('status', 'dockwrkr.command.status')
    self.addCommand('exec', 'dockwrkr.command.exec')
    self.addCommand('stats', 'dockwrkr.command.stats')
    return self

  def getShellOptions(self, optparser):
    optparser.add_option("-f", dest="configFile", help="Override default config file")
    optparser.add_option("-d", dest="debug", help="Activate debugging output", default=False, action="store_true")
    optparser.add_option("-y", dest="assumeYes", help="Assume yes when prompted", default=False, action="store_true")
    return optparser

  def getUsage(self):
    return "dockwrkr [options] COMMAND [command-options]"

  def getHelpTitle(self):
    version = getPackageVersion()
    return "Docker container composition (version: %s)" % version

  def initCommand(self, command):
    core = Core()
    if self.getOption('assumeYes'):
      command.assumeYes = True

    if self.getOption('configFile') and os.path.isfile(self.getOption('configFile')):
      logging.debug("CONF %s" % self.getOption('configFile'))
      core.configFile = self.getOption('configFile')

    if command.autoInitCore: 
      core.initialize().catch(self.exitError)

    command.core = core
    return command

def cli():
  args = sys.argv
  args.pop(0)

  prog = DockwrkrCLI()
  prog.execute(args)



from dockwrkr import (Command)

class Stats(Command):

  def getShellOptions(self, optparser):
    return optparser

  def getUsage(self):
    return "dockwrkr stats [options] CONTAINER..."

  def getHelpTitle(self):
    return "Output live stats for the listed containers"

  def main(self):
    containers = self.args
    return self.core.stats(self.args) \
      .catch(self.exitError)



import os
import sys
import logging
import shutil
import shlex
import subprocess
from subprocess import call, Popen, check_output, CalledProcessError
from dockwrkr.monads import *
from dockwrkr.exceptions import (FileSystemError, ShellCommandError, UserInterruptError)
from threading import Thread

# pylint: disable=W0232

logger = logging.getLogger(__name__)

class Shell(object):

  @staticmethod
  def printConfirm(msg, assumeYes=False):
    if assumeYes:
      return OK(True)

    logger.info(msg)
    try:
      res = raw_input('Proceed? [N/y] ')
      if res.lower().startswith('y'):
        logger.info('... proceeding')
        return OK(True)
      return Fail(UserInterruptError(message="User interrupted."))
    except KeyboardInterrupt as err:
      return Fail(UserInterruptError(message="User interrupted."))


  @staticmethod
  def call(cmd, cwd=None, shell=True):
    logger.debug("COMMAND[%s]: %s", cwd, cmd)
    try:
      returncode = call(cmd, shell=shell, cwd=cwd)
      if returncode == 0:
        return OK(None)
      else:
        return Fail(ShellCommandError(code=returncode))
    except CalledProcessError as err:
      return Fail(ShellCommandError(code=err.returncode, message=err.output, stdout=err.output))
    except KeyboardInterrupt as err:
      logger.info("CTRL-C Received...Exiting.")
      return Fail(UserInterruptError(message="User interrupted."))

  @staticmethod
  def command(cmd, shell=False, cwd=None):
    logger.debug("COMMAND: %s", cmd)
    try:
      out = check_output(shlex.split(cmd), shell=shell, cwd=cwd)
      return OK({'code':0, 'stdout':out.strip(), 'stderr':''})
    except CalledProcessError as err:
      return Fail(ShellCommandError(code=err.returncode, message=err.output, stdout=err.output))

  @staticmethod
  def procCommand(cmd, cwd=None, shell=False):
    try:
      logger.debug("COMMAND: %s", cmd)
      proc = Popen(shlex.split(cmd), shell=shell, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
      pout, perr = proc.communicate()
      if proc.returncode == 0:
        return OK({"code": proc.returncode, "stdout": pout, "stderr": perr})
      else:
        return Fail(ShellCommandError(code=proc.returncode, message=pout, stdout=pout, stderr=perr))
    except KeyboardInterrupt:
      logger.info("CTRL-C Received...Exiting.")
      return Fail(UserInterruptError(message="User interrupted."))

  @staticmethod
  def streamCommand(cmd, cwd=None, shell=False, stream=False):
    try:
      logger.debug("COMMAND: %s", cmd)
      proc = Popen(shlex.split(cmd), shell=shell, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
      stdout = ''
      stderr = ''
      while True:
        d = proc.stdout.read(1)
        if d != '':
          stdout += d
        if stream:
          sys.stdout.write(d)
          sys.stdout.flush()

        de = proc.stderr.read(1)
        if de != '':
          stderr += de
        if stream:
          sys.stderr.write(de)
          sys.stderr.flush()

        if d == '' and de == '' and proc.poll() is not None:
          break

      if proc.returncode == 0:
        return OK({"code": proc.returncode, "stdout": stdout, "stderr": stderr})
      else:
        return Fail(ShellCommandError(code=proc.returncode, message=stdout, stdout=stdout, stderr=stderr))
    except OSError as err:
      err.strerror = "Error running '%s': %s" % (cmd, err.strerror)
      return Fail(err)
    except KeyboardInterrupt:
      logger.info("CTRL-C Received...Exiting.")
      return Fail(UserInterruptError(message="User interrupted."))

  @staticmethod
  def chmod(path, mode):
    try:
      os.chmod(path, mode)
      return OK(path)
    except Exception as err:
      return Fail(ShellCommandError(code=1, message="Failed to chmod %s: %s" % (path,err)))
    
  @staticmethod
  def makeDirectory(path, mode=0750):
    if not os.path.exists(path):
      try:
        os.makedirs(path, mode)
        return OK(path)
      except Exception as err:
        return Fail(ShellCommandError(code=1, message="Failed to create %s: %s" % (path,err)))
    return OK(path)

  @staticmethod
  def copyFile(src, dst):
    return Try.attempt(lambda: shutil.copy(src, dst))
    
  @staticmethod
  def pathExists(path):
    return OK(path) if os.path.exists(path) else Fail(FileSystemError("Path %s does not exist." % path))

  @staticmethod
  def rmFile(path):
    if os.path.isdir(path):
      return Fail(FileSystemError("%s is a directory.") % path)
    return OK(os.remove(path))

  @staticmethod
  def nukeDirectory(path):
    try:
      if path and path is not '/' and path is not 'C:\\':
        shutil.rmtree(path)
      return OK(None)
    except Exception as err:
      return Fail(ShellCommandError(code=1, message="Failed to rmtree: %s: %s" % (path,err)))



import sys
import logging
import logging.config
from dockwrkr.monads import *

class StdoutFilter(logging.Filter):
  def __init__(self, level):
    self._level = level
    logging.Filter.__init__(self)
  def filter(self, rec):
    return rec.levelno < self._level

class LevelLogFormatter(logging.Formatter):
  def format(self, record):
    if record.levelno == logging.DEBUG:
      record.msg = '[%s] %s' % (record.levelname, record.msg)
    return super(LevelLogFormatter , self).format(record)

LOG_SETTINGS = {
  'version': 1,
  'disable_existing_loggers': False, 
  'root': {
    'level': 'NOTSET',
  },
  'loggers': {
    'dockwrkr': {
      'level': 'NOTSET',
      'qualname': 'dockwrkr',
      'handlers': ['stdout', 'stderr'],
      'propagate': False
    }
  },
  'handlers': {
    'console': {
      'level': 'NOTSET',
      'formatter': 'levelformatter',
      'class': 'logging.StreamHandler',
      'stream': 'ext://sys.stdout',
    },
    'stdout': {
      'level': 'INFO',
      'formatter': 'levelformatter',
      'class': 'logging.StreamHandler',
      'stream': 'ext://sys.stdout',
      'filters': ['StdoutFilter'],
    },
    'stderr': {
      'level': 'WARNING',
      'formatter': 'levelformatter',
      'class': 'logging.StreamHandler',
      'stream': 'ext://sys.stderr'
    }
  },
  'filters': {
    'StdoutFilter': {
      '()': StdoutFilter,
      'level': logging.WARNING
    }
  },
  'formatters': {
    'levelformatter': { 
      '()': LevelLogFormatter, 'format': '%(message)s' },
    'stdout': { 'format': '%(message)s' },
    'stderr': { 'format': '%(message)s' }
  }
}

logging.config.dictConfig(LOG_SETTINGS)

logger = logging.getLogger("dockwrkr")


def ddebug(msg, *args, **kwargs):
  return dlog(logging.DEBUG, msg, *args, **kwargs)

def dinfo(msg, *args, **kwargs):
  return dlog(logging.INFO, msg, *args, **kwargs)

def dwarning(msg, *args, **kwargs):
  return dlog(logging.WARNING, msg, *args, **kwargs)

def dcritical(msg, *args, **kwargs):
  return dlog(logging.CRITICAL, msg, *args, **kwargs)

def dexception(msg, *args, **kwargs):
  return dlog(logging.EXCEPTION, msg, *args, **kwargs)

def dlog(level, msg, *args, **kwargs):
  if args:
    msg = msg % args
  def deferredLog(data=None, *args, **kwargs):
    logger.log(level, msg)
    return OK(data)
  return deferredLog

